import numpy as np

import gdsfactory as gf


@gf.cell
def coupler_bent_half(
    gap: float = 0.200,
    radius: float = 26,
    length: float = 8.6,
    width1: float = 0.400,
    width2: float = 0.400,
    length_straight: float = 10,
    cross_section: str = "xs_sc",
) -> gf.Component:
    """Returns Broadband SOI curved / straight directional coupler.

    Args:
        gap: gap.
        radius: radius coupling.
        length: coupler_length.
        width1: width1.
        width2: width2.
        length_straight: input and output straight length.
        cross_section: cross_section.
    """
    R1 = radius + (width1 + gap) / 2
    R2 = radius - (width2 + gap) / 2
    alpha = round(np.rad2deg(length / (2 * radius)), 4)
    beta = alpha

    c = gf.Component()

    xs = gf.get_cross_section(cross_section)
    xs1 = xs.copy(radius=R1, width=width1)
    xs2 = xs.copy(radius=R2, width=width2)

    outer_bend = c << gf.components.bend_circular(
        angle=np.round(-alpha, 3), cross_section=xs1, add_pins=False
    )

    inner_bend = c << gf.components.bend_circular(
        angle=-alpha, cross_section=xs2, add_pins=False
    )

    outer_bend.movey(+(width1 + gap) / 2)
    inner_bend.movey(-(width2 + gap) / 2)

    outer_straight = c << gf.components.straight(
        length=length, cross_section=xs1, npoints=100, add_pins=False
    )

    inner_straight = c << gf.components.straight(
        length=length, cross_section=xs2, npoints=100, add_pins=False
    )

    outer_straight.connect(port="o1", destination=outer_bend.ports["o2"])
    inner_straight.connect(port="o1", destination=inner_bend.ports["o2"])

    outer_exit_bend = c << gf.components.bend_circular(
        angle=alpha, cross_section=xs1, add_pins=False
    )

    inner_exit_bend_down = c << gf.components.bend_circular(
        angle=-beta, cross_section=xs2, add_pins=False
    )

    inner_exit_bend_up = c << gf.components.bend_circular(
        angle=alpha + beta, cross_section=xs2, add_pins=False
    )

    outer_exit_bend.connect(port="o1", destination=outer_straight.ports["o2"])
    inner_exit_bend_down.connect(port="o1", destination=inner_straight.ports["o2"])
    inner_exit_bend_up.connect(port="o1", destination=inner_exit_bend_down.ports["o2"])

    inner_exit_straight = c << gf.components.straight(
        length=length_straight,
        cross_section=gf.cross_section.cross_section(width=width2),
        npoints=100,
    )
    inner_exit_straight.connect(port="o1", destination=inner_exit_bend_up.ports["o2"])

    outer_exit_straight = c << gf.components.straight(
        length=abs(inner_exit_straight.ports["o2"].x - outer_exit_bend.ports["o2"].x),
        cross_section=gf.cross_section.cross_section(width=width1),
        npoints=100,
    )
    outer_exit_straight.connect(port="o1", destination=outer_exit_bend.ports["o2"])

    c.add_port("o1", port=outer_bend.ports["o1"])
    c.add_port("o2", port=inner_bend.ports["o1"])
    c.add_port("o3", port=outer_exit_straight.ports["o2"])
    c.add_port("o4", port=inner_exit_straight.ports["o2"])

    c.absorb(outer_exit_bend)
    c.absorb(inner_exit_bend_down)
    c.absorb(inner_exit_bend_up)
    c = c.flatten_invalid_refs()
    return c


@gf.cell
def coupler_bent(
    gap: float = 0.200,
    radius: float = 26,
    length: float = 8.6,
    width1: float = 0.400,
    width2: float = 0.400,
    length_straight: float = 10,
    cross_section: str = "xs_sc",
) -> gf.Component:
    """Returns Broadband SOI curved / straight directional coupler.
    based on: https://doi.org/10.1038/s41598-017-07618-6

    Args:
        gap: gap.
        radius: radius coupling.
        length: coupler_length.
        width1: width1.
        width2: width2.
        length_straight: input and output straight length.
        cross_section: cross_section.
    """
    c = gf.Component()
    xs = gf.get_cross_section(cross_section)

    right_half = c << coupler_bent_half(
        gap=gap,
        radius=radius,
        length=length,
        width1=width1,
        width2=width2,
        length_straight=length_straight,
        cross_section=cross_section,
    )
    left_half = c << coupler_bent_half(
        gap=gap,
        radius=radius,
        length=length,
        width1=width1,
        width2=width2,
        length_straight=length_straight,
        cross_section=cross_section,
    )

    left_half.mirror_x()
    left_half.connect(port="o1", destination=right_half.ports["o1"])
    c = c.flatten()

    c.add_port("o1", port=left_half.ports["o3"])
    c.add_port("o2", port=left_half.ports["o4"])
    c.add_port("o3", port=right_half.ports["o3"])
    c.add_port("o4", port=right_half.ports["o4"])

    xs.add_pins(c)
    return c


if __name__ == "__main__":
    # c = coupler_bent_half()
    c = coupler_bent()
    c.show(show_ports=False)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler_ring import coupler_ring as coupler_ring_function
from gdsfactory.components.straight import straight
from gdsfactory.typings import ComponentFactory, CrossSectionSpec


@gf.cell
def ring_single(
    gap: float = 0.2,
    radius: float = 10.0,
    length_x: float = 4.0,
    length_y: float = 0.6,
    coupler_ring: ComponentFactory = coupler_ring_function,
    bend: ComponentFactory = bend_euler,
    cross_section: CrossSectionSpec = "xs_sc",
    **kwargs,
) -> gf.Component:
    """Returns a single ring.

    ring coupler (cb: bottom) connects to two vertical straights (sl: left, sr: right),
    two bends (bl, br) and horizontal straight (wg: top)

    Args:
        gap: gap between for coupler.
        radius: for the bend and coupler.
        length_x: ring coupler length.
        length_y: vertical straight length.
        coupler_ring: ring coupler spec.
        bend: 90 degrees bend spec.
        cross_section: cross_section spec.
        **kwargs: cross_section settings.

    .. code::

          bl-st-br
          |      |
          sl     sr length_y
          |      |
         --==cb==-- gap

          length_x
    """
    gap = gf.snap.snap_to_grid2x(gap)

    xs = gf.get_cross_section(cross_section)
    radius = radius or xs.radius
    cross_section = xs.copy(radius=radius, **kwargs)

    c = gf.Component()
    cb = c << coupler_ring(
        bend=bend,
        gap=gap,
        radius=radius,
        length_x=length_x,
        cross_section=cross_section,
    )
    sy = straight(length=length_y, cross_section=cross_section)
    b = bend(cross_section=cross_section)
    sx = straight(length=length_x, cross_section=cross_section)

    sl = sy.ref()
    sr = sy.ref()
    st = sx.ref()

    if length_y > 0:
        c.add(sl)
        c.add(sr)

    if length_x > 0:
        c.add(st)

    bl = c << b
    br = c << b

    sl.connect(port="o1", destination=cb.ports["o2"])
    bl.connect(port="o2", destination=sl.ports["o2"])

    st.connect(port="o2", destination=bl.ports["o1"])
    br.connect(port="o2", destination=st.ports["o1"])
    sr.connect(port="o1", destination=br.ports["o1"])
    sr.connect(port="o2", destination=cb.ports["o3"])

    c.add_port("o2", port=cb.ports["o4"])
    c.add_port("o1", port=cb.ports["o1"])
    return c


if __name__ == "__main__":
    # c = ring_single(layer=(2, 0), cross_section_factory=gf.cross_section.pin, width=1)
    # c = ring_single(width=2, gap=1, layer=(2, 0), radius=7, length_y=1)
    # print(c.ports)

    # c = gf.routing.add_fiber_array(ring_single)
    # c = ring_single(cross_section="xs_rc", width=2)
    c = ring_single(length_y=0, length_x=0)
    c.get_netlist()
    c.show(show_ports=True)

    # cc = gf.add_pins(c)
    # print(c.settings)
    # print(c.settings)
    # cc.show(show_ports=True)



from __future__ import annotations

from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.components.via import via
from gdsfactory.components.via_stack import via_stack
from gdsfactory.cross_section import Section, xs_rc
from gdsfactory.typings import ComponentSpec, CrossSectionFactory, LayerSpec

cross_section_rib = partial(
    gf.cross_section.strip,
    sections=(Section(width=2 * 2.425, layer="SLAB90", name="slab"),),
)
cross_section_pn = partial(
    gf.cross_section.pn,
    width_doping=2.425,
    width_slab=2 * 2.425,
    layer_via="VIAC",
    width_via=0.5,
    layer_metal="M1",
    width_metal=0.5,
)
heater_vias = partial(
    via_stack,
    size=(0.5, 0.5),
    layers=("M1", "M2"),
    vias=(
        partial(
            via,
            layer="VIAC",
            size=(0.1, 0.1),
            spacing=(0.2, 0.2),
            enclosure=0.1,
        ),
        partial(
            via,
            layer="VIA1",
            size=(0.1, 0.1),
            spacing=(0.2, 0.2),
            enclosure=0.1,
        ),
    ),
)


@gf.cell
def ring_double_pn(
    add_gap: float = 0.3,
    drop_gap: float = 0.3,
    radius: float = 5.0,
    doping_angle: float = 85,
    cross_section: CrossSectionFactory = xs_rc,
    pn_cross_section: CrossSectionFactory = cross_section_pn,
    doped_heater: bool = True,
    doped_heater_angle_buffer: float = 10,
    doped_heater_layer: LayerSpec = "NPP",
    doped_heater_width: float = 0.5,
    doped_heater_waveguide_offset: float = 2.175,
    heater_vias: ComponentSpec = heater_vias,
    **kwargs,
) -> gf.Component:
    """Returns add-drop pn ring with optional doped heater.

    Args:
        add_gap: gap to add waveguide.
        drop_gap: gap to drop waveguide.
        radius: for the bend and coupler.
        doping_angle: angle in degrees representing portion of ring that is doped.
        length_x: ring coupler length.
        length_y: vertical straight length.
        cross_section: cross_section spec for non-PN doped rib waveguide sections.
        pn_cross_section: cross section of pn junction.
        doped_heater: boolean for if we include doped heater or not.
        doped_heater_angle_buffer: angle in degrees buffering heater from pn junction.
        doped_heater_layer: doping layer for heater.
        doped_heater_width: width of doped heater.
        doped_heater_waveguide_offset: distance from the center of the ring waveguide to the center of the doped heater.
        heater_vias: components specifications for heater vias
        kwargs: cross_section settings.

    """

    add_gap = gf.snap.snap_to_grid(add_gap, grid_factor=2)
    drop_gap = gf.snap.snap_to_grid(drop_gap, grid_factor=2)
    c = gf.Component()

    pn_cross_section = gf.get_cross_section(pn_cross_section)
    pn_cross_section = pn_cross_section.copy(**kwargs)
    cross_section = gf.get_cross_section(cross_section)
    cross_section = cross_section.copy(**kwargs)

    undoping_angle = 180 - doping_angle

    add_waveguide_path = gf.Path()
    add_waveguide_path.append(
        gf.path.straight(length=2 * radius * np.sin(np.pi / 360 * undoping_angle))
    )
    add_waveguide = c << add_waveguide_path.extrude(cross_section=cross_section)
    add_waveguide.x = 0
    add_waveguide.y = 0

    drop_waveguide_path = gf.Path()
    drop_waveguide_path.append(
        gf.path.straight(length=2 * radius * np.sin(np.pi / 360 * undoping_angle))
    )
    drop_waveguide = c << drop_waveguide_path.extrude(cross_section=cross_section)
    drop_waveguide.x = 0

    doped_path = gf.Path()
    doped_path.append(gf.path.arc(radius=radius, angle=-doping_angle))
    undoped_path = gf.Path()
    undoped_path.append(gf.path.arc(radius=radius, angle=undoping_angle))
    left_doped_ring_ref = c << doped_path.extrude(cross_section=pn_cross_section)
    right_doped_ring_ref = c << doped_path.extrude(cross_section=pn_cross_section)
    bottom_undoped_ring_ref = c << undoped_path.extrude(cross_section=cross_section)
    top_undoped_ring_ref = c << undoped_path.extrude(cross_section=cross_section)

    bottom_undoped_ring_ref.rotate(-undoping_angle / 2)
    bottom_undoped_ring_ref.ymin = (
        add_waveguide.ymin + add_waveguide.ports["o1"].width + add_gap
    )
    bottom_undoped_ring_ref.x = add_waveguide.x

    left_doped_ring_ref.connect("o1", bottom_undoped_ring_ref.ports["o1"])
    right_doped_ring_ref.connect("o2", bottom_undoped_ring_ref.ports["o2"])
    top_undoped_ring_ref.connect("o2", left_doped_ring_ref.ports["o2"])

    drop_waveguide.y = (
        2 * radius
        + add_gap
        + drop_gap
        + add_waveguide.ports["o1"].width / 2
        + top_undoped_ring_ref.ports["o1"].width
        + drop_waveguide.ports["o1"].width / 2
    )

    if doped_heater:
        heater_radius = radius - doped_heater_waveguide_offset
        heater_path = gf.Path()
        heater_path.append(
            gf.path.arc(
                radius=heater_radius, angle=undoping_angle - doped_heater_angle_buffer
            )
        )

        top_heater_ref = c << heater_path.extrude(width=0.5, layer=doped_heater_layer)
        top_heater_ref.rotate(180 - (undoping_angle - doped_heater_angle_buffer) / 2)
        top_heater_ref.x = add_waveguide.x
        top_heater_ref.ymax = drop_waveguide.y - (
            doped_heater_waveguide_offset + doped_heater_width / 2 + drop_gap
        )

        top_left_heater_via = c << heater_vias()
        top_left_heater_via.rotate(top_heater_ref.ports["o2"].orientation)

        deltax = -abs(top_heater_ref.ports["o2"].x - top_left_heater_via.ports["e3"].x)
        deltay = abs(top_heater_ref.ports["o2"].y - top_left_heater_via.ports["e3"].y)
        top_left_heater_via.move((deltax, deltay))

        top_right_heater_via = c << heater_vias()
        top_right_heater_via.rotate(top_heater_ref.ports["o1"].orientation)

        deltax = abs(top_heater_ref.ports["o1"].x - top_right_heater_via.ports["e3"].x)
        deltay = abs(top_heater_ref.ports["o1"].y - top_right_heater_via.ports["e3"].y)
        top_right_heater_via.move((deltax, deltay))

        bottom_heater_ref = c << heater_path.extrude(
            width=0.5, layer=doped_heater_layer
        )
        bottom_heater_ref.rotate(-(undoping_angle - doped_heater_angle_buffer) / 2)
        bottom_heater_ref.x = add_waveguide.x
        bottom_heater_ref.ymin = (
            doped_heater_waveguide_offset + doped_heater_width / 2 + add_gap
        )

        bottom_left_heater_via = c << heater_vias()
        bottom_left_heater_via.rotate(bottom_heater_ref.ports["o1"].orientation)

        deltax = -abs(
            bottom_heater_ref.ports["o1"].x - bottom_left_heater_via.ports["e3"].x
        )
        deltay = abs(
            bottom_heater_ref.ports["o1"].y - bottom_left_heater_via.ports["e3"].y
        )
        bottom_left_heater_via.move((deltax, deltay))

        bottom_right_heater_via = c << heater_vias()
        bottom_right_heater_via.rotate(bottom_heater_ref.ports["o2"].orientation)

        deltax = abs(
            bottom_heater_ref.ports["o2"].x - bottom_right_heater_via.ports["e3"].x
        )
        deltay = abs(
            bottom_heater_ref.ports["o2"].y - bottom_right_heater_via.ports["e3"].y
        )
        bottom_right_heater_via.move((deltax, deltay))

    c.add_port("o1", port=add_waveguide.ports["o1"])
    c.add_port("o2", port=add_waveguide.ports["o2"])
    c.add_port("o3", port=drop_waveguide.ports["o2"])
    c.add_port("o4", port=drop_waveguide.ports["o1"])
    return c


if __name__ == "__main__":
    c = ring_double_pn()
    c.show()



"""Polygon

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""

from __future__ import annotations

import gdstk
import numpy as np
import shapely as sp

from gdsfactory.component_layout import _GeometryHelper, _parse_move, _simplify
from gdsfactory.snap import snap_to_grid


class Polygon(gdstk.Polygon, _GeometryHelper):
    """Polygonal geometric object.

    Args:
        points: array-like[N][2] Coordinates of the vertices of the Polygon.
        gds_layer: int GDSII layer of the Polygon.
        gds_datatype: int GDSII datatype of the Polygon.
    """

    def __repr__(self) -> str:
        """Returns path points."""
        return f"Polygon(layer=({self.layer}, {self.datatype}), points={self.points})"

    def __init__(self, points, layer):
        from gdsfactory.pdk import get_layer

        layer, datatype = get_layer(layer)
        super().__init__(points, layer, datatype)

    @property
    def bbox(self):
        """Returns the bounding box of the Polygon."""
        return self.bounding_box()

    def rotate(
        self, angle: float = 45, center: tuple[float, float] = (0, 0)
    ) -> Polygon:
        """Rotates a Polygon by the specified angle.

        Args:
            angle: Angle to rotate the Polygon in degrees.
            center: Midpoint of the Polygon.
        """
        super().rotate(angle=angle * np.pi / 180, center=center)
        return self

    def move(
        self,
        origin: tuple[float, float] = (0, 0),
        destination: tuple[float, float] | None = None,
        axis: str | None = None,
    ) -> Polygon:
        """Moves elements of the Device from the origin point to the
        destination. Both origin and destination can be 1x2 array-like, Port,
        or a key corresponding to one of the Ports in this device.

        Args:
            origin: Origin point of the move.
            destination : Destination point of the move.
            axis: {'x', 'y'} Direction of move.

        """
        dx, dy = _parse_move(origin, destination, axis)

        super().translate(dx, dy)
        return self

    def mirror(self, p1=(0, 1), p2=(0, 0)) -> Polygon:
        """Mirrors a Polygon across the line formed between the two
        specified points. ``points`` may be input as either single points
        [1,2] or array-like[N][2], and will return in kind.

        Args:
            p1: First point of the line.
            p2: Second point of the line.
        """
        return super().mirror(p1, p2)

    def simplify(self, tolerance: float = 1e-3) -> Polygon:
        """Removes points from the polygon but does not change the polygon
        shape by more than `tolerance` from the original. Uses the
        Ramer-Douglas-Peucker algorithm.

        Args:
            tolerance: float
                Tolerance value for the simplification algorithm.  All points that
                can be removed without changing the resulting polygon by more than
                the value listed here will be removed. Also known as `epsilon` here
                https://en.wikipedia.org/wiki/Ramer%E2%80%93Douglas%E2%80%93Peucker_algorithm
        """

        points = _simplify(self.points, tolerance=tolerance)
        return Polygon(points, (self.layer, self.datatype))

    def snap(self, grid_factor: int = 1, nm: int | None = None) -> Polygon:
        """Returns new polygon snap points to grid"""
        points = snap_to_grid(self.points, grid_factor=grid_factor, nm=nm)
        return Polygon(points, (self.layer, self.datatype))

    def to_shapely(self) -> sp.Polygon:
        import shapely as sp

        return sp.Polygon(self.points)

    @classmethod
    def from_shapely(cls, polygon: sp.Polygon, layer) -> Polygon:
        from gdsfactory.pdk import get_layer

        layer, datatype = get_layer(layer)
        points_on_grid = snap_to_grid(polygon.exterior.coords)
        polygon = Polygon(points_on_grid, (layer, datatype))
        return polygon


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component("demo")
    p0 = Polygon(zip((-8, 6, 7, 9), (-6, 8, 17, 5)), layer="WG")
    p1 = sp.Polygon(zip((-8, 6, 7, 9), (-6, 8, 17, 5)))
    p2 = p1.buffer(1)
    p3 = c.add_polygon(p2, layer=1)
    p4 = p3.simplify(tolerance=0.1)
    p5 = c.add_polygon(p4, layer=2)
    p5.mirror()
    c.show()



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler_ring import coupler_ring as _coupler_ring
from gdsfactory.components.straight import straight
from gdsfactory.components.via_stack import via_stack_heater_mtop
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, Float2

via_stack_heater_mtop_mini = partial(via_stack_heater_mtop, size=(4, 4))


@gf.cell
def ring_single_heater(
    gap: float = 0.2,
    radius: float = 10.0,
    length_x: float = 4.0,
    length_y: float = 0.6,
    coupler_ring: ComponentSpec = _coupler_ring,
    bend: ComponentSpec = bend_euler,
    cross_section_waveguide_heater: CrossSectionSpec = "xs_sc_heater_metal",
    cross_section: CrossSectionSpec = "xs_sc",
    via_stack: ComponentSpec = via_stack_heater_mtop_mini,
    port_orientation: float | None = None,
    via_stack_offset: Float2 = (0, 0),
) -> gf.Component:
    """Returns a single ring with heater on top.

    ring coupler (cb: bottom) connects to two vertical straights (sl: left, sr: right),
    two bends (bl, br) and horizontal straight (wg: top)

    Args:
        gap: gap between for coupler.
        radius: for the bend and coupler.
        length_x: ring coupler length.
        length_y: vertical straight length.
        coupler_ring: ring coupler function.
        bend: 90 degrees bend function.
        cross_section_waveguide_heater: for heater.
        cross_section: for regular waveguide.
        via_stack: for heater to routing metal.
        port_orientation: for electrical ports to promote from via_stack.
        via_stack_offset: x,y offset for via_stack.

    .. code::

          bl-st-br
          |      |
          sl     sr length_y
          |      |
         --==cb==-- gap

          length_x
    """
    gap = gf.snap.snap_to_grid(gap, grid_factor=2)

    coupler_ring = gf.get_component(
        coupler_ring,
        bend=bend,
        gap=gap,
        radius=radius,
        length_x=length_x,
        cross_section=cross_section,
        cross_section_bend=cross_section_waveguide_heater,
    )

    straight_side = straight(
        length=length_y,
        cross_section=cross_section_waveguide_heater,
    )
    straight_top = straight(
        length=length_x,
        cross_section=cross_section_waveguide_heater,
    )

    bend = gf.get_component(
        bend,
        radius=radius,
        cross_section=cross_section_waveguide_heater,
    )

    c = gf.Component()
    cb = c << coupler_ring
    sl = c << straight_side
    sr = c << straight_side
    bl = c << bend
    br = c << bend
    st = c << straight_top
    # st.mirror(p1=(0, 0), p2=(1, 0))

    sl.connect(port="o1", destination=cb.ports["o2"])
    bl.connect(port="o2", destination=sl.ports["o2"])

    st.connect(port="o2", destination=bl.ports["o1"])
    br.connect(port="o2", destination=st.ports["o1"])
    sr.connect(port="o1", destination=br.ports["o1"])
    sr.connect(port="o2", destination=cb.ports["o3"])

    c.add_port("o2", port=cb.ports["o4"])
    c.add_port("o1", port=cb.ports["o1"])

    via = gf.get_component(via_stack)
    c1 = c << via
    c2 = c << via
    c1.xmax = -length_x / 2 + cb.x - via_stack_offset[0]
    c2.xmin = +length_x / 2 + cb.x + via_stack_offset[0]
    c1.movey(via_stack_offset[1])
    c2.movey(via_stack_offset[1])

    p1 = c1.get_ports_list(orientation=port_orientation)
    p2 = c2.get_ports_list(orientation=port_orientation)
    valid_orientations = {p.orientation for p in via.ports.values()}

    if not p1:
        raise ValueError(
            f"No ports found for port_orientation {port_orientation} in {valid_orientations}"
        )

    c.add_ports(p1, prefix="l_")
    c.add_ports(p2, prefix="r_")
    return c


if __name__ == "__main__":
    c = ring_single_heater()
    c.show(show_subports=False)



"""LVS demo."""

from __future__ import annotations

import gdsfactory as gf


@gf.cell
def pads_correct(pad=gf.components.pad, cross_section="xs_m3") -> gf.Component:
    """Returns 2 pads connected with metal wires."""
    c = gf.Component()

    xs = gf.get_cross_section(cross_section)
    layer = gf.get_layer(xs.layer)

    pad = gf.get_component(pad)
    tl = c << pad
    bl = c << pad

    tr = c << pad
    br = c << pad

    tl.move((0, 300))
    br.move((500, 0))
    tr.move((500, 500))

    c.add_label("tl", position=tl.center, layer=layer)
    c.add_label("tr", position=tr.center, layer=layer)
    c.add_label("br", position=br.center, layer=layer)
    c.add_label("bl", position=bl.center, layer=layer)

    ports1 = [bl.ports["e3"], tl.ports["e3"]]
    ports2 = [br.ports["e1"], tr.ports["e1"]]
    routes = gf.routing.get_bundle(ports1, ports2, cross_section=cross_section)

    for route in routes:
        c.add(route.references)

    return c


@gf.cell
def pads_shorted(pad=gf.components.pad, cross_section="xs_m3") -> gf.Component:
    """Returns 2 pads connected with metal wires."""

    c = gf.Component()
    pad = gf.components.pad()
    xs = gf.get_cross_section(cross_section)
    layer = gf.get_layer(xs.layer)

    tl = c << pad
    bl = c << pad

    tr = c << pad
    br = c << pad

    tl.move((0, 300))
    br.move((500, 0))
    tr.move((500, 500))

    c.add_label("tl", position=tl.center, layer=layer)
    c.add_label("tr", position=tr.center, layer=layer)
    c.add_label("br", position=br.center, layer=layer)
    c.add_label("bl", position=bl.center, layer=layer)

    ports1 = [bl.ports["e3"], tl.ports["e3"]]
    ports2 = [br.ports["e1"], tr.ports["e1"]]
    routes = gf.routing.get_bundle(ports1, ports2, cross_section=cross_section)

    for route in routes:
        c.add(route.references)

    route = gf.routing.get_route(
        bl.ports["e2"], tl.ports["e4"], cross_section=cross_section
    )
    c.add(route.references)
    return c


if __name__ == "__main__":
    c = pads_correct()
    c.show(show_ports=True)
    gdspath = c.write_gds()

    import kfactory as kf

    lib = kf.kcell.KCLayout()
    lib.read(filename=str(gdspath))
    c = lib[0]

    l2n = kf.kdb.LayoutToNetlist(c.begin_shapes_rec(0))
    for l_idx in c.kcl.layer_indices():
        l2n.connect(l2n.make_layer(l_idx, f"layer{l_idx}"))
    l2n.extract_netlist()
    print(l2n.netlist().to_s())



from __future__ import annotations

import warnings
from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.components.via import via1, via2, viac
from gdsfactory.typings import ComponentSpec, Float2, Floats, LayerSpec, LayerSpecs


@gf.cell
def via_stack(
    size=(11.0, 11.0),
    layers: LayerSpecs = ("M1", "M2", "M3"),
    layer_offsets: Floats | None = None,
    vias: tuple[ComponentSpec | None, ...] | None = (via1, via2, None),
    layer_port: LayerSpec | None = None,
    correct_size: bool = True,
) -> Component:
    """Rectangular via array stack.

    You can use it to connect different metal layers or metals to silicon.
    You can use the naming convention via_stack_layerSource_layerDestination
    contains 4 ports (e1, e2, e3, e4)

    also know as Via array
    http://www.vlsi-expert.com/2017/12/vias.html

    spacing = via.info['spacing']
    enclosure = via.info['enclosure']

    Args:
        size: of the layers.
        layers: layers on which to draw rectangles.
        layer_offsets: Optional offsets for each layer with respect to size.
            positive grows, negative shrinks the size.
        vias: vias to use to fill the rectangles.
        layer_port: if None assumes port is on the last layer.
        correct_size: if True, if the specified dimensions are too small it increases
            them to the minimum possible to fit a via.
    """
    width_m, height_m = size
    a = width_m / 2
    b = height_m / 2

    layers = layers or []
    layer_offsets = layer_offsets or [0] * len(layers)

    elements = {len(layers), len(layer_offsets), len(vias)}
    if len(elements) > 1:
        warnings.warn(
            f"Got {len(layers)} layers, {len(layer_offsets)} layer_offsets, {len(vias)} vias",
            stacklevel=3,
        )

    if layers:
        layer_port = layer_port or layers[-1]

    c = Component()
    c.height = height_m
    c.info["size"] = (float(size[0]), float(size[1]))
    c.info["layer"] = layer_port

    for layer, offset in zip(layers, layer_offsets):
        size_m = (width_m + 2 * offset, height_m + 2 * offset)
        if layer == layer_port:
            ref = c << compass(size=size_m, layer=layer, port_type="electrical")
            c.add_ports(ref.ports)
        else:
            ref = c << compass(size=size_m, layer=layer, port_type="electrical")

    vias = vias or []
    for via, offset in zip(vias, layer_offsets):
        if via is not None:
            width, height = size
            width += 2 * offset
            height += 2 * offset
            via = gf.get_component(via)
            w, h = via.info["size"]
            enclosure = via.info["enclosure"]
            pitch_x, pitch_y = via.info["spacing"]

            min_width = w + enclosure
            min_height = h + enclosure

            if (
                min_width > width
                and correct_size
                or min_width <= width
                and min_height > height
                and correct_size
            ):
                warnings.warn(
                    f"Changing size from ({width}, {height}) to ({min_width}, {min_height}) to fit a via!",
                    stacklevel=3,
                )
                width = max(min_width, width)
                height = max(min_height, height)
            elif min_width > width or min_height > height:
                raise ValueError(f"size {size} is too small to fit a {(w, h)} um via")

            nb_vias_x = abs(width - w - 2 * enclosure) / pitch_x + 1
            nb_vias_y = abs(height - h - 2 * enclosure) / pitch_y + 1

            nb_vias_x = int(np.floor(nb_vias_x)) or 1
            nb_vias_y = int(np.floor(nb_vias_y)) or 1
            ref = c.add_array(
                via, columns=nb_vias_x, rows=nb_vias_y, spacing=(pitch_x, pitch_y)
            )
            if ref.xsize + enclosure > width or ref.ysize + enclosure > height:
                warnings.warn(
                    f"size = {size} for layer {layer} violates min enclosure"
                    f" {enclosure} for via {via.name!r}",
                    stacklevel=3,
                )

            a = width / 2
            b = height / 2
            cw = (width - (nb_vias_x - 1) * pitch_x - w) / 2
            ch = (height - (nb_vias_y - 1) * pitch_y - h) / 2
            x0 = -a + cw + w / 2
            y0 = -b + ch + h / 2
            ref.move((x0, y0))

    return c


@gf.cell
def via_stack_circular(
    radius: float = 10.0,
    angular_extent: float = 45,
    center_angle: float = 0,
    width: float = 5.0,
    layers: LayerSpecs = ("M1", "M2", "M3"),
    vias: tuple[ComponentSpec | None, ...] = (via1, via2),
    layer_port: LayerSpec | None = None,
) -> Component:
    """Circular via array stack.

    FIXME! does not work.

    Constructs a circular via array stack. It does so
    by stacking rectangular via stacks offset by a small amount
    along the specified circumference.

    Args:
        radius: of the via stack (center).
        angular_extent: of the via stack.
        center_angle: of the via stack.
        width: of the via stack.
        layers: layers to draw
        vias: vias to use to fill the rectangles.
        layer_port: if None assumes port is on the last layer.
    """

    # We basically just want to place rectangular via stacks
    # stacked with a little bit of an offset
    c = gf.Component()

    layers = layers or []

    if layers:
        layer_port = layer_port or layers[-1]

    init_angle = (center_angle - angular_extent / 2) * np.pi / 180
    end_angle = (center_angle + angular_extent / 2) * np.pi / 180

    init_angle = init_angle % (2 * np.pi)
    if init_angle > np.pi:
        init_angle = init_angle - 2 * np.pi

    end_angle = end_angle % (2 * np.pi)
    if end_angle > np.pi:
        end_angle = end_angle - 2 * np.pi

    # We do this via-centric: we figure out the min spacing between vias,
    # and from that figure out all the metal dimensions
    # This will of course fail if no via information is provided,
    # but why would you instantiate a ViaStack without any via?

    for level, via in enumerate(vias):
        if via is None:
            continue

        metal_bottom = layers[level]
        metal_top = layers[level + 1]
        via = gf.get_component(via)

        w, h = via.info["size"]
        g = via.info["enclosure"]
        pitch_x, pitch_y = via.info["spacing"]

        nb_vias_x = (width - w - 2 * g) / pitch_x + 1
        nb_vias_x = int(np.floor(nb_vias_x)) or 1

        size_metal = (width, h + 2 * g)

        # Now start placing via lines at each angle starting from
        # the initial angle until we reach the end angle
        ang = init_angle

        while _smaller_angle(ang, ang, end_angle):
            pos = radius * np.array((np.cos(ang), np.sin(ang)))

            ref = c.add_array(
                via, columns=nb_vias_x, rows=1, spacing=(pitch_x, pitch_y)
            )
            ref.center = pos

            # Place top and bottom metal
            for metal in [metal_top, metal_bottom]:
                met = c << gf.components.rectangle(size=size_metal, layer=metal)
                met.center = pos

            # Let's see if we can do something different
            x, y = pos

            if x > 0:
                new_y = y + pitch_y
                mult = 1
            else:
                new_y = y - pitch_y
                mult = -1

            if new_y > radius:
                new_y = y - pitch_y
                assert new_y < radius
                new_x = -1 * np.sqrt(np.power(radius, 2) - np.power(new_y, 2))
            elif new_y < -radius:
                new_y = y + pitch_y
                assert new_y > -radius
                new_x = np.sqrt(np.power(radius, 2) - np.power(new_y, 2))

            else:
                new_x = mult * np.sqrt(np.power(radius, 2) - np.power(new_y, 2))

            if np.isnan(new_x):
                print(radius)
                print(new_y)
                print(np.power(radius, 2) - np.power(new_y, 2))
            assert not np.isnan(new_x)
            ang = np.arctan2(new_y, new_x)

    return c


def _smaller_angle(angle, angle1, angle2):
    """Returns False if angle is outside the
     bounds of the arc angle defined between
     angle 1 and angle2.

    But it does so assuming that angle1 and angle2 are between [-pi, pi]
    and that we are trying to fill an arc
    """

    if angle2 >= 0 and angle1 >= 0:
        if angle2 > angle1:
            return angle < angle2
        # Convert angle to 0, 2pi and see if out of bounds
        angle = angle + 2 * np.pi * (angle < 0)
        return not (angle2 < angle < angle1)

    elif angle2 < 0 and angle1 < 0:
        return angle < angle2 if angle2 > angle1 else not (angle2 < angle < angle1)
    else:
        if angle2 < 0 and angle > 0 or angle2 >= 0 and angle < 0:
            return True
        else:
            return angle < angle2


@gf.cell
def via_stack_from_rules(
    size: Float2 = (1.2, 1.2),
    layers: LayerSpecs = ("M1", "M2", "M3"),
    layer_offsets: tuple[float, ...] | None = None,
    vias: tuple[ComponentSpec | None, ...] | None = (via1, via2),
    via_min_size: tuple[Float2, ...] = ((0.2, 0.2), (0.2, 0.2)),
    via_min_gap: tuple[Float2, ...] = ((0.1, 0.1), (0.1, 0.1)),
    via_min_enclosure: Float2 = (0.15, 0.25),
    layer_port: LayerSpec | None = None,
) -> Component:
    """Rectangular via array stack, with optimized dimension for vias.

    Uses inclusion, minimum width, and minimum spacing rules to place the maximum number of individual vias,
    each with maximum via area.

    Args:
        size: of the layers, len(size).
        layers: layers on which to draw rectangles.
        layer_offsets: Optional offsets for each layer with respect to size.
            positive grows, negative shrinks the size.
        vias: list of base via components to modify.
        via_min_size: via minimum x, y dimensions.
        via_min_gap: via minimum x, y distances.
        via_min_enclosure: via minimum inclusion into connecting layers.
        layer_port: if None assumes port is on the last layer.
    """
    width, height = size
    a = width / 2
    b = height / 2

    layers = layers or []

    if layers:
        layer_port = layer_port or layers[-1]

    c = Component()
    c.height = height
    c.info["size"] = (float(size[0]), float(size[1]))
    c.info["layer"] = layer_port

    layer_offsets = layer_offsets or [0] * len(layers)

    for layer, offset in zip(layers, layer_offsets):
        size = (width + 2 * offset, height + 2 * offset)
        if layer == layer_port:
            ref = c << compass(size=size, layer=layer, port_type="electrical")
            c.add_ports(ref.ports)
        else:
            ref = c << compass(size=size, layer=layer, port_type="electrical")

    vias = vias or []
    c.info["vias"] = []
    for current_via, min_size, min_gap, min_enclosure in zip(
        vias, via_min_size, via_min_gap, via_min_enclosure
    ):
        if current_via is not None:
            # Optimize via
            via = gf.get_component(
                optimized_via(current_via, size, min_size, min_gap, min_enclosure)
            )
            c.info["vias"].append(via.info)

            w, h = via.info["size"]
            g = via.info["enclosure"]
            pitch_x, pitch_y = via.info["spacing"]

            nb_vias_x = (width - w - 2 * g) / pitch_x + 1
            nb_vias_y = (height - h - 2 * g) / pitch_y + 1

            nb_vias_x = int(np.floor(nb_vias_x)) or 1
            nb_vias_y = int(np.floor(nb_vias_y)) or 1
            ref = c.add_array(
                via, columns=nb_vias_x, rows=nb_vias_y, spacing=(pitch_x, pitch_y)
            )

            cw = (width - (nb_vias_x - 1) * pitch_x - w) / 2
            ch = (height - (nb_vias_y - 1) * pitch_y - h) / 2
            x0 = -a + cw + w / 2
            y0 = -b + ch + h / 2
            ref.move((x0, y0))

    return c


def optimized_via(
    base_via: ComponentSpec = "VIAC",
    size: tuple[float, float] = (11.0, 11.0),
    min_via_size: tuple[float, float] = (0.3, 0.3),
    min_via_gap: tuple[float, float] = (0.1, 0.1),
    min_via_enclosure: float = 0.2,
) -> Component:
    """Given a target total inclusion size, returns an optimized dimension for the via.

    Uses inclusion, minimum width, and minimum spacing rules to place the maximum number of individual vias, with maximum via area.

    Arguments:
        base_via: to modify.
        size: of the target enclosing medium.
        min_via_size: minimum size the vias can take.
        min_via_gap: minimum distance between vias.
        min_via_enclosure: minimum distance between edge of enclosing medium and nearest via edge.
    """
    via_size = [0, 0]
    for dim in [0, 1]:
        via_area = size[dim] - 2 * min_via_enclosure + min_via_gap[dim]
        num_vias = int(via_area / (min_via_size[dim] + min_via_gap[dim]))
        try:
            via_size[dim] = float(via_area / num_vias) - min_via_gap[dim]
        except ZeroDivisionError as e:
            raise RuntimeError(
                "Cannot fit vias with specified minimum dimensions in provided space."
            ) from e

    return partial(
        base_via,
        size=via_size,
        gap=min_via_gap,
        spacing=None,
        enclosure=min_via_enclosure,
    )


def test_via_stack_from_rules() -> None:
    # Check that vias are generated with larger than min dimensions if possible
    size = (1.2, 1.2)
    layers = ("M1", "M2", "M3")
    vias = (via1, via2)
    via_min_size = ((0.2, 0.2), (0.2, 0.2))
    via_min_gap = ((0.1, 0.1), (0.15, 0.15))
    via_min_enclosure = (0.1, 0.1)

    c = gf.get_component(
        via_stack_from_rules(
            size=size,
            layers=layers,
            vias=vias,
            via_min_size=via_min_size,
            via_min_gap=via_min_gap,
            via_min_enclosure=via_min_enclosure,
        )
    )

    assert c.info["vias"][0]["size"][0] > via_min_size[0][0]
    assert c.info["vias"][0]["size"][1] > via_min_size[0][1]
    assert (
        c.info["vias"][0]["spacing"][0]
        == via_min_gap[0][0] + c.info["vias"][0]["size"][0]
    )


via_stack_m1_m3 = partial(
    via_stack,
    layers=("M1", "M2", "M3"),
    vias=(via1, via2, None),
)

via_stack_slab_m3 = partial(
    via_stack,
    layers=("SLAB90", "M1", "M2", "M3"),
    vias=(viac, via1, via2, None),
)
via_stack_slab_m2 = partial(
    via_stack,
    layers=("SLAB90", "M1", "M2"),
    vias=(viac, via1, None),
)
via_stack_npp_m1 = partial(
    via_stack,
    layers=("WG", "NPP", "M1"),
    vias=(None, None, viac),
)
via_stack_slab_npp_m3 = partial(
    via_stack,
    layers=("SLAB90", "NPP", "M1"),
    vias=(None, None, viac),
)
via_stack_heater_mtop = via_stack_heater_m3 = partial(
    via_stack, layers=("HEATER", "M2", "M3"), vias=(None, via1, via2)
)


if __name__ == "__main__":
    c = via_stack()
    c.show()
    # c = gf.pack([via_stack_slab_m3, via_stack_heater_mtop])[0]

    # c = gf.Component("offgrid_demo")
    # v1 = c << via_stack_slab_m3()
    # v2 = c << via_stack_slab_m3()
    # v2.x = 20.0005
    # c.show()

    # c2 = gf.Component()
    # c21 = c2 << c
    # c22 = c2 << c
    # c22.x = 20.0005 + 30
    # c2.show()

    # c = via_stack_heater_mtop(layer_offsets=(0, 1, 2))
    # c = via_stack_circular()
    # c = via_stack_m1_m3(size=(4.5, 4.5))
    # print(c.to_dict())
    # c.show(show_ports=True)

    # c = via_stack_from_rules()
    # c.show(show_ports=True)

    # c = via_stack_circular(
    #     radius=20.0,
    #     angular_extent=300,
    #     center_angle=0,
    #     width=5.0,
    #     layers=("M1", "M2", "M3"),
    #     vias=(via1, via2),
    #     layer_port=None,
    # )
    # c.show(show_ports=True)



from __future__ import annotations

import numpy as np

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import LayerSpec


@cell
def optimal_hairpin(
    width: float = 0.2,
    pitch: float = 0.6,
    length: float = 10,
    turn_ratio: float = 4,
    num_pts: int = 50,
    layer: LayerSpec = (1, 0),
) -> Component:
    """Returns an optimally-rounded hairpin geometry, with a 180 degree turn \
    on the right end of the polygon connected to two prongs extending towards \
    ports on the left end.

    based on phidl.geometry

    Args:
        width: Width of the hairpin leads.
        pitch: Distance between the two hairpin leads. Must be greater than width.
        length: Length of the hairpin from the connectors to the opposite end of the curve.
        turn_ratio: int or float
            Specifies how much of the hairpin is dedicated to the 180 degree turn.
            A turn_ratio of 10 will result in 20% of the hairpin being comprised of the turn.
        num_pts: Number of points constituting the 180 degree turn.
        layer: Specific layer(s) to put polygon geometry on.

    Notes:
        Hairpin pitch must be greater than width.

        Optimal structure from https://doi.org/10.1103/PhysRevB.84.174510
        Clem, J., & Berggren, K. (2011). Geometry-dependent critical currents in
        superconducting nanocircuits. Physical Review B, 84(17), 1–27.
    """
    # ==========================================================================
    #  Create the basic geometry
    # ==========================================================================
    a = (pitch + width) / 2
    y = -(pitch - width) / 2
    x = -pitch
    dl = width / (num_pts * 2)
    n = 0

    # Get points of ideal curve from conformal mapping
    # TODO This is an inefficient way of finding points that you need
    xpts = [x]
    ypts = [y]
    while (y < 0) & (n < 1e6):
        s = x + 1j * y
        w = np.sqrt(1 - np.exp(np.pi * s / a))
        wx = np.real(w)
        wy = np.imag(w)
        wx = wx / np.sqrt(wx**2 + wy**2)
        wy = wy / np.sqrt(wx**2 + wy**2)
        x = x + wx * dl
        y = y + wy * dl
        xpts.append(x)
        ypts.append(y)
        n += 1
    ypts[-1] = 0  # Set last point be on the x=0 axis for sake of cleanliness
    ds_factor = len(xpts) // num_pts
    xpts = xpts[::-ds_factor]
    xpts = xpts[::-1]  # This looks confusing, but it's just flipping the arrays around
    ypts = ypts[::-ds_factor]
    ypts = ypts[::-1]  # so the last point is guaranteed to be included when downsampled

    # Add points for the rest of meander
    xpts.append(xpts[-1] + turn_ratio * width)
    ypts.append(0)
    xpts.append(xpts[-1])
    ypts.append(-a)
    xpts.append(xpts[0])
    ypts.append(-a)
    xpts.append(max(xpts) - length)
    ypts.append(-a)
    xpts.append(xpts[-1])
    ypts.append(-a + width)
    xpts.append(xpts[0])
    ypts.append(ypts[0])

    xpts = snap_to_grid(xpts)
    ypts = snap_to_grid(ypts)

    # ==========================================================================
    #  Create a blank device, add the geometry, and define the ports
    # ==========================================================================
    c = Component()
    c.add_polygon([xpts, ypts], layer=layer)
    c.add_polygon([xpts, -ypts], layer=layer)

    xports = min(xpts)
    yports = -a + width / 2
    c.add_port(
        name="e1", center=(xports, -yports), width=width, orientation=180, layer=layer
    )
    c.add_port(
        name="e2", center=(xports, yports), width=width, orientation=180, layer=layer
    )
    return c


if __name__ == "__main__":
    c = optimal_hairpin()
    c.show(show_ports=True)



"""Cell decorator for functions that return a Component."""
from __future__ import annotations

import functools
import hashlib
import inspect
from collections.abc import Callable
from dataclasses import dataclass
from functools import wraps
from typing import Any, TypeVar

import toolz
from pydantic import BaseModel, validate_call

from gdsfactory.component import Component, name_counters
from gdsfactory.config import CONF
from gdsfactory.name import clean_name, get_name_short
from gdsfactory.serialization import clean_dict, clean_value_name

CACHE: dict[str, Component] = {}

INFO_VERSION = 2

_F = TypeVar("_F", bound=Callable)


class CellReturnTypeError(ValueError):
    pass


def clear_cache() -> None:
    """Clears Component CACHE."""
    global CACHE

    CACHE = {}
    name_counters.clear()


def print_cache() -> None:
    for k in CACHE:
        print(k)


def get_source_code(func: Callable) -> str:
    if isinstance(func, functools.partial):
        source = inspect.getsource(func.func)
    elif isinstance(func, toolz.functoolz.Compose):
        source = inspect.getsource(func.first)
    elif callable(func):
        source = inspect.getsource(func)
    else:
        raise ValueError(f"{func!r} needs to be callable")
    return source


class Settings(BaseModel):
    name: str
    function_name: str | None = None
    module: str | None = None

    info: dict[str, Any] = {}  # derived properties (length, resistance)
    info_version: int = INFO_VERSION

    full: dict[str, Any] = {}
    changed: dict[str, Any] = {}
    default: dict[str, Any] = {}

    child: dict[str, Any] | None = None


def cell_without_validator(func: _F) -> _F:
    """Decorator for Component functions.

    Similar to cell decorator, this one does not validate_arguments using
    type annotations

    I recommend using @cell instead
    """

    @functools.wraps(func)
    def _cell(*args, **kwargs):
        from gdsfactory.pdk import get_active_pdk

        active_pdk = get_active_pdk()
        cell_decorator_settings = active_pdk.cell_decorator_settings

        ports_off_grid = kwargs.pop("ports_off_grid", CONF.ports_off_grid)
        ports_not_manhattan = kwargs.pop(
            "ports_not_manhattan", CONF.ports_not_manhattan
        )

        with_hash = kwargs.pop("with_hash", cell_decorator_settings.with_hash)
        autoname = kwargs.pop("autoname", cell_decorator_settings.autoname)
        name = kwargs.pop("name", cell_decorator_settings.name)
        cache = kwargs.pop("cache", cell_decorator_settings.cache)
        flatten = kwargs.pop("flatten", cell_decorator_settings.flatten)
        info = kwargs.pop("info", {})
        prefix = kwargs.pop(
            "prefix",
            func.__name__
            if cell_decorator_settings.prefix is None
            else cell_decorator_settings.prefix,
        )
        max_name_length = kwargs.pop(
            "max_name_length", cell_decorator_settings.max_name_length
        )
        include_module = kwargs.pop(
            "include_module", cell_decorator_settings.include_module
        )

        sig = inspect.signature(func)
        args_as_kwargs = dict(zip(sig.parameters.keys(), args))
        args_as_kwargs.update(kwargs)

        default = {
            p.name: p.default
            for p in sig.parameters.values()
            if p.default != inspect._empty
        }

        changed = args_as_kwargs
        full = default.copy()
        full.update(**args_as_kwargs)

        default2 = default.copy()
        changed2 = changed.copy()

        # list of default args as strings
        default_args_list = [
            f"{key}={clean_value_name(default2[key])}" for key in sorted(default.keys())
        ]
        # list of explicitly passed args as strings
        passed_args_list = [
            f"{key}={clean_value_name(changed2[key])}" for key in sorted(changed.keys())
        ]

        # get only the args which are explicitly passed and different from defaults
        # if any args were different from default, append a hash of those args.
        # else, keep only the base name
        # print(named_args_string)

        if active_pdk.cell_decorator_settings.naming_style == "updk":
            full_args_list = [
                f"{key}={clean_value_name(full[key])}" for key in sorted(full.keys())
            ]
            named_args_string = ",".join(full_args_list)
            name = f"{prefix}:{named_args_string}" if named_args_string else prefix
            name = clean_name(name, allowed_characters=[":", ".", "="])

        elif active_pdk.cell_decorator_settings.naming_style == "default":
            changed_arg_set = set(passed_args_list).difference(default_args_list)
            changed_arg_list = sorted(changed_arg_set)
            named_args_string = "_".join(changed_arg_list)
            if changed_arg_list or include_module:
                if include_module and changed_arg_list:
                    named_args_module_string = f"{named_args_string}_{func.__module__}"
                elif include_module:
                    named_args_module_string = func.__module__
                elif changed_arg_list:
                    named_args_module_string = named_args_string

                named_args_module_string = (
                    hashlib.md5(named_args_module_string.encode()).hexdigest()[:8]
                    if with_hash
                    or len(named_args_module_string) > 28
                    or "'" in named_args_module_string
                    or "{" in named_args_module_string
                    else named_args_module_string
                )
                name_signature = clean_name(f"{prefix}_{named_args_module_string}")
            else:
                name_signature = prefix

            # filter the changed dictionary to only keep entries which have truly changed
            changed_arg_names = [carg.split("=")[0] for carg in changed_arg_list]
            changed = {k: changed[k] for k in changed_arg_names}
            name = name or name_signature

        else:
            raise ValueError('naming_style must be "default" or "updk"')

        name = get_name_short(name, max_name_length=max_name_length)
        default_decorator = active_pdk.default_decorator if active_pdk else None
        decorator = kwargs.pop("decorator", default_decorator)

        if (
            "args" not in sig.parameters
            and "kwargs" not in sig.parameters
            and "settings" not in sig.parameters
        ):
            for key in kwargs:
                if key not in sig.parameters.keys():
                    raise TypeError(
                        f"{func.__name__!r}() got invalid argument {key!r}\n"
                        f"valid arguments are {list(sig.parameters.keys())}"
                    )

        if cache and name in CACHE:
            # print(f"CACHE LOAD {name} {func.__name__}({named_args_string})")
            return CACHE[name]

        # print(f"BUILD {name} {func.__name__}({named_args_string})")
        if not callable(func):
            raise ValueError(
                f"{func!r} is not callable! @cell decorator is only for functions"
            )

        component = func(*args, **kwargs)

        if ports_off_grid in ("warn", "error"):
            component.assert_ports_on_grid(error_type=ports_off_grid)
        if ports_not_manhattan in ("warn", "error"):
            component.assert_ports_manhattan(error_type=ports_off_grid)

        if flatten:
            component = component.flatten()

        # if the component is already in the cache, but under a different alias,
        # make sure we use a copy, so we don't run into mutability errors
        if id(component) in [id(v) for v in CACHE.values()]:
            component = component.copy()

        metadata_child = (
            dict(component.child.settings) if hasattr(component, "child") else None
        )

        if not isinstance(component, Component):
            raise CellReturnTypeError(
                f"function {func.__name__!r} return type = {type(component)}",
                "make sure that functions with @cell decorator return a Component",
            )

        if metadata_child and component._get_child_name:
            component_name = f"{metadata_child.get('name')}_{name}"
            component_name = get_name_short(
                component_name, max_name_length=max_name_length
            )
        else:
            component_name = name

        if autoname and not hasattr(component, "imported_gds"):
            component.name = component_name

        info.update(**cell_decorator_settings.info)
        component.info.update(**info)

        if not hasattr(component, "imported_gds"):
            component.settings = Settings(
                name=component_name,
                function_name=func.__name__,
                module=func.__module__,
                changed=clean_dict(changed),
                default=clean_dict(default),
                full=clean_dict(full),
                info=component.info,
                child=metadata_child,
            )
            component.__doc__ = func.__doc__

        if decorator:
            if not callable(decorator):
                raise ValueError(f"decorator = {type(decorator)} needs to be callable")
            component_new = decorator(component)
            # if component_new is not component:
            #     component_new.name = name
            component = component_new or component

        component.lock()
        CACHE[name] = component
        return component

    return _cell


def cell(func: _F) -> _F:
    """Decorator for Component functions.

    Wraps cell_without_validator
    Validates type annotations with pydantic.

    Implements a cache so that if a component has already been build
    it will return the component from the cache directly.
    This avoids creating two exact Components that have the same name.

    When decorate your functions with @cell you get:

    - cache: avoids creating duplicated Components.
    - name: names Components uniquely name based on parameters.
    - metadata: adds Component.metadata with default, changed and full Args.

    Note the cell decorator does not take any arguments.
    Keyword Args are applied the resulting Component.

    Keyword Args:
        autoname (bool): True renames Component based on args and kwargs.
            True by default.
        name (str): Optional name.
        cache (bool): returns Component from the CACHE if it already exists.
            Avoids having duplicated cells with the same name.
            If False overrides CACHE creates a new Component.
        flatten (bool): False by default. True flattens component hierarchy.
        info: updates Component.info dict.
        prefix (str): name_prefix, defaults to function name.
        max_name_length (int): truncates name beyond some characters with a hash.
        decorator (Callable): function to apply to Component.


    A decorator is a function that runs over a function, so when you do.

    .. code::

        import gdsfactory as gf

        @gf.cell
        def mzi_with_bend():
            c = gf.Component()
            mzi = c << gf.components.mzi()
            bend = c << gf.components.bend_euler()
            return c

    it’s equivalent to

    .. code::

        def mzi_with_bend():
            c = gf.Component()
            mzi = c << gf.components.mzi()
            bend = c << gf.components.bend_euler(radius=radius)
            return c

        mzi_with_bend_decorated = gf.cell(mzi_with_bend)

    """

    return cell_without_validator(validate_call(func))


def declarative_cell(cls: type[Any]) -> Callable[..., Component]:
    """Decorator to turn a dataclass into a cell. Work in progress."""
    cls = dataclass(cls)

    @wraps(cls)
    def cell(*args, **kwargs):
        decl = cls(*args, **kwargs)

        sig = inspect.signature(cls)
        args_as_kwargs = dict(zip(sig.parameters.keys(), args))
        args_as_kwargs.update(kwargs)

        args_list = [
            f"{key}={clean_value_name(args_as_kwargs[key])}"
            for key in sorted(args_as_kwargs.keys())
        ]
        named_args_string = "_".join(args_list)
        component_name = clean_name(f"{cls.__name__}_{named_args_string}")
        if component_name in CACHE:
            return CACHE[component_name]

        decl.instances()
        comp = Component()
        comp.name = component_name

        for k, c in vars(decl).items():
            if not isinstance(c, Component):
                continue
            ref = comp << c
            setattr(comp, k, ref)
            setattr(decl, k, ref)
        for p1, p2 in decl.connections():
            p1.reference.connect(p1.name, p2.reference.ports[p2.name])
        for name, p in decl.ports().items():
            comp.add_port(name, port=p.reference.ports[p.name])
        CACHE[component_name] = comp
        return comp

    return cell


@cell
def wg(length: int = 3, layer: tuple[int, int] = (1, 0)) -> Component:
    """Dummy component for testing."""
    c = Component()
    width = 0.5
    w = width / 2
    c.add_polygon([(0, -w), (length, -w), (length, w), (0, w)], layer=layer)
    c.add_port(name="o1", center=[0, 0], width=width, orientation=180, layer=layer)
    c.add_port(name="o2", center=[length, 0], width=width, orientation=0, layer=layer)
    return c


@cell
def wg2(wg1=wg):
    """Dummy component for testing."""
    c = Component()
    w = wg1()
    w1 = c << w
    w1.rotate(90)
    c.copy_child_info(w)
    c.add_ports(w1.ports)
    return c


def test_set_name() -> None:
    c = wg(length=3, name="hi_there")
    assert c.name == "hi_there", c.name


@cell
def demo(length: int = 3, wg_width: float = 0.5) -> Component:
    """Demo Dummy cell."""
    c = Component()
    w = length
    h = wg_width
    points = [
        [-w / 2.0, -h / 2.0],
        [-w / 2.0, h / 2],
        [w / 2, h / 2],
        [w / 2, -h / 2.0],
    ]
    c.add_polygon(points)
    return c


def test_names() -> None:
    name_base = demo().name
    assert name_base.split("_")[0] == "demo", name_base

    demo2 = functools.partial(demo, length=3)
    c1 = demo2(length=3)
    c2 = demo(length=3)
    assert c1.name == c2.name, "{c1.name} != {c2.name}"

    c1 = demo(length=3, wg_width=0.5).name
    c2 = demo(wg_width=0.5, length=3).name
    assert c1 == c2, f"{c1} != {c2}"

    name_with_prefix = demo(prefix="hi").name
    assert name_with_prefix.split("_")[0] == "hi", name_with_prefix

    name_args = demo(3).name
    name_kwargs = demo(length=3).name
    assert name_args == name_kwargs, name_with_prefix


@cell
def straight_with_pins(**kwargs) -> Component:
    import gdsfactory as gf

    c = gf.Component()
    ref = c << gf.components.straight()
    c.add_ports(ref.ports)
    gf.add_pins.add_pins(c)
    return c


def test_hashes() -> None:
    import gdsfactory as gf

    c = gf.components.mzi()
    names1 = {i.name for i in c.get_dependencies()}
    c = gf.components.mzi()
    names2 = {i.name for i in c.get_dependencies()}
    assert names1 == names2


if __name__ == "__main__":
    import gdsfactory as gf

    # c = gf.c.straight(length=3, info=dict(polarization="te", wavelength=1.55))
    c = gf.c.straight()
    gf.clear_cache()
    c = gf.c.straight()
    print(c.name)
    # c.pprint()
